<?php

namespace app\Controllers;

use app\Core\Request;
use app\Core\Response;
use app\Forms\Field;
use app\Forms\Number;
use app\Forms\Required;
use app\Forms\Validation;
use app\Models\Dorm;
use app\Models\Media;
use app\Models\Owners;
use app\Models\Reviews;
use app\Utils\FileManager;
use app\Utils\Toast;

class DormController extends Controller
{
  public function getList()
  {
    $page = Request::getPage();
    $cities = Dorm::getCities();
    $ids = Request::getQuery()["ids"] ?? null;
    $sort = Request::getQuery()["sort"] ?? "dorm_id DESC";
    $city = Request::getQuery()["city"] ?? "";
    $price = Request::getQuery()["price"] ?? "";
    $priceValues = [
      "price <= 1000000",
      "price >= 1000000 AND price <= 2000000",
      "price >= 2000000"
    ];
    if (!in_array($sort, ["dorm_id DESC", "price DESC", "price ASC", "rate"])) {
      $sort = "dorm_id DESC";
    }
    if (!in_array($city, $cities)) {
      $city = "";
    }
    if (!in_array($price, $priceValues)) {
      $price = "";
    }
    if ($ids !== null) {
      $ids = explode(",", $ids);
      $data = Dorm::getInIds($ids, $page, $sort, $city, $price);
      $data["page"] = $page;
      $data["cities"] = $cities;
      return $data;
    } else {
      $data = Dorm::getAllFiltered($page, $sort, Request::getQuery()["q"] ?? "", $city, $price);
    }
    $data["page"] = $page;
    $data["cities"] = $cities;
    return $data;
  }

  public function list()
  {
    $data = $this->getList();
    if (Request::isJSON()) {
      Response::json($data);
      return;
    }
    if (isset(Request::getQuery()["list-only"])) {
      $this->render("dorm/list", $data);
      return;
    }
    $this->render('index', $data);
  }

  public function view($params)
  {
    $dormId = $params["dormId"];
    $dorm = Dorm::findById($dormId);
    $medias = Media::findByDormId($dormId);
    $owner = Owners::findById($dorm->owner_id);
    $reviews = Reviews::findByDormId($dorm->dorm_id);

    if (Request::getMethod() === "POST") {
      Response::redirect("/dorms/{$dorm->dorm_id}/edit");
      return;
    }

    $data = [
      "dorm" => $dorm,
      "medias" => $medias,
      "owner" => $owner,
    ];

    if (Request::isJSON()) {
      Response::json($data);
      return;
    }
    $data["reviews"] = $reviews;

    $this->render("dorm/view", $data);
  }


  public function create()
  {
    $form = $this->getDormForm();

    if (Request::getMethod() === "POST") {
      if ($form->validate(Request::getBody())) {
        $form->data["name"] = $form->data["dorm_name"];
        unset($form->data["dorm_name"]);
        $dorm = Dorm::toModel($form->data);
        $dorm->save();
        Response::redirect("/dorms/{$dorm->dorm_id}/media");
        return;
      }
      var_dump($form->getFields()["errors"]);
      exit();
    }
    $owners = Owners::all("name DESC");
    $this->render("dorm/create", [
      "owners" => $owners,
      ...$form->getFields()
    ]);
  }


  public function media($params)
  {
    $dormId  = $params["dormId"];
    if (Request::getMethod() === "POST") {
      $files = Request::getFiles("medias");
      if ($files !== null) {
        $this->uploadMedia($files, $dormId);
        Response::redirect("/dorms/{$dormId}");
        return;
      }
    }
    $dorm = Dorm::findById($dormId);
    $medias = Media::findByDormId($dormId);

    $this->render("dorm/media", [
      "dorm" => $dorm,
      "medias" => $medias
    ]);
  }

  public function deleteMedia($params)
  {
    $form = new Validation([
      [new Field("media_id"), [new Number()]]
    ]);
    if ($form->validate(Request::getBody())) {
      $media = Media::findById($form->data["media_id"]);
      $dormId = $media->dorm_id;
      $media->delete();
    }
    Response::redirect("/dorms/{$dormId}/media");
  }

  private function uploadMedia($files, $dorm_id)
  {
    foreach ($files["name"] as $index => $name) {
      $filePath = FileManager::getPathFor($name, $files["type"][$index]);
      if ($filePath == false) {
        continue;
      }
      $media = new Media();
      $media->dorm_id = $dorm_id;
      $media->alt_text = $name;
      $media->type = explode("/", $files["type"][$index])[0] === "image" ? "photo" : "video";
      $media->endpoint = $filePath;
      $media->save();
      FileManager::uploadToPublic($files["tmp_name"][$index], $filePath);
    }
  }

  public function edit($params)
  {

    $dormId = $params["dormId"];
    $form = $this->getDormForm();
    if (Request::getMethod() === "POST") {
      if ($form->validate(Request::getBody())) {
        $form->data["name"] = $form->data["dorm_name"];
        unset($form->data["dorm_name"]);
        $dorm = Dorm::toModel($form->data);
        $dorm->dorm_id = $dormId;
        $dorm->save();
        Response::redirect("/dorms/{$dorm->dorm_id}");
        return;
      }
    } else {
      $dorm = Dorm::findById($dormId);
      $this->setDormFormData($form, $dorm);
    }
    $owners = Owners::all("name DESC");

    $this->render("dorm/create", [
      "owners" => $owners,
      "dormId" => $dormId,
      ...$form->getFields()
    ]);
  }

  public function delete($params)
  {
    $dormId = $params["dormId"];
    Dorm::deleteById($dormId);
    Toast::success("Kos berhasil dihapus", true);
    Response::redirect("/");
  }

  private function setDormFormData(Validation $form, Dorm $dorm)
  {
    $form->data["dorm_name"] = $dorm->name;
    $form->data["address"] = $dorm->address;
    $form->data["price"] = $dorm->price;
    $form->data["description"] = $dorm->description;
    $form->data["city"] = $dorm->city;
    $form->data["owner_id"] = $dorm->owner_id;
  }

  private function getDormForm()
  {
    return new Validation([
      [new Field("dorm_name", "Nama Kos"), [new Required()]],
      [new Field("address", "Alamat"), [new Required()]],
      [new Field("price", "Harga"), [new Required(), new Number()]],
      [new Field("description", "Deskripsi"), [new Required()]],
      [new Field("city", "Kota"), [new Required()]],
      [new Field("owner_id", "Pemilik"), [new Required(), new Number()]]
    ]);
  }
}
