<?php

namespace app\Core;

use app\Controllers\Controller;

class Router
{
  protected $routes = [];
  protected $dynamicRoutes = [];

  public function get(string $route, array $middlewares, $controller, $action)
  {
    $vars = $this->isDynamic($route);
    if ($vars === false) {
      $this->addStaticRoute($route, $middlewares, $controller, $action, "GET");
    } else {
      $routeRegex = $this->getRouteRegex($route);
      $this->addDynamicRoute($routeRegex, $vars, $middlewares, $controller, $action, "GET");
    }
  }

  public function post(string $route, array $middlewares, $controller, $action)
  {
    $vars = $this->isDynamic($route);
    if ($vars === false) {
      $this->addStaticRoute($route, $middlewares, $controller, $action, "POST");
    } else {
      $routeRegex = $this->getRouteRegex($route);
      $this->addDynamicRoute($routeRegex, $vars, $middlewares, $controller, $action, "POST");
    }
  }

  public function put(string $route, array $middlewares, $controller, $action)
  {
    $vars = $this->isDynamic($route);
    if ($vars === false) {
      $this->addStaticRoute($route, $middlewares, $controller, $action, "PUT");
    } else {
      $routeRegex = $this->getRouteRegex($route);
      $this->addDynamicRoute($routeRegex, $vars, $middlewares, $controller, $action, "PUT");
    }
  }

  public function delete(string $route, array $middlewares, $controller, $action)
  {
    $vars = $this->isDynamic($route);
    if ($vars === false) {
      $this->addStaticRoute($route, $middlewares, $controller, $action, "DELETE");
    } else {
      $routeRegex = $this->getRouteRegex($route);
      $this->addDynamicRoute($routeRegex, $vars, $middlewares, $controller, $action, "DELETE");
    }
  }

  public function all(string $route, array $middlewares, $controller, $action)
  {
    $vars = $this->isDynamic($route);
    if ($vars === false) {
      $this->addStaticRoute($route, $middlewares, $controller, $action, "GET");
      $this->addStaticRoute($route, $middlewares, $controller, $action, "POST");
      $this->addStaticRoute($route, $middlewares, $controller, $action, "PUT");
      $this->addStaticRoute($route, $middlewares, $controller, $action, "DELETE");
    } else {
      $routeRegex = $this->getRouteRegex($route);
      $this->addDynamicRoute($routeRegex, $vars, $middlewares, $controller, $action, "GET");
      $this->addDynamicRoute($routeRegex, $vars, $middlewares, $controller, $action, "POST");
      $this->addDynamicRoute($routeRegex, $vars, $middlewares, $controller, $action, "PUT");
      $this->addDynamicRoute($routeRegex, $vars, $middlewares, $controller, $action, "DELETE");
    }
  }

  public function methods(array $methods, string $route, array $middlewares, $controller, $action)
  {
    $vars = $this->isDynamic($route);
    if ($vars === false) {
      foreach ($methods as $method) {
        $this->addStaticRoute($route, $middlewares, $controller, $action, $method);
      }
    } else {
      $routeRegex = $this->getRouteRegex($route);
      foreach ($methods as $method) {
        $this->addDynamicRoute($routeRegex, $vars, $middlewares, $controller, $action, $method);
      }
    }
  }

  protected function addStaticRoute(string $route, array &$middlewares, $controller, $action, $method = "GET")
  {
    if (array_key_exists($route, $this->routes) === false) {
      $this->routes[$route] = [$method => [$controller, $middlewares, $action]];
    } else {
      $this->routes[$route][$method] = [$controller, $middlewares, $action];
    }
  }

  protected function addDynamicRoute(string $routeRegex, &$vars, array &$middlewares, $controller, $action, $method = "GET")
  {
    if (array_key_exists($routeRegex, $this->dynamicRoutes) === false) {
      $this->dynamicRoutes[$routeRegex] = [$method => [$controller, $middlewares, $action, $vars]];
    } else {
      $this->dynamicRoutes[$routeRegex][$method] = [$controller, $middlewares, $action, $vars];
    }
  }

  public function dispatch()
  {
    if ($this->dispatchStatic()) {
      return;
    }
    if ($this->dispatchDynamic()) {
      return;
    }
    Controller::notFound();
  }

  private function dispatchStatic()
  {
    $uri = Request::getPath();
    if (array_key_exists($uri, $this->routes) === false) {
      return false;
    }

    $uriHandler = $this->routes[$uri];
    if (array_key_exists(Request::getMethod(), $uriHandler) === false) {
      return false;
    }
    [$controllerClass, $middlewares, $action] = $uriHandler[Request::getMethod()];
    if ($this->runMiddlewares($middlewares)) {
      $controller = new $controllerClass();
      $controller->$action();
    }
    return true;
  }

  private function dispatchDynamic()
  {
    $uri = Request::getPath();
    foreach ($this->dynamicRoutes as $routeRegex => $routeHandler) {
      if (preg_match($routeRegex, $uri, $matches)) {
        if (array_key_exists(Request::getMethod(), $routeHandler) === false) {
          return false;
        }
        [$controllerClass, $middlewares, $action, $vars] = $routeHandler[Request::getMethod()];
        $controller = new $controllerClass();
        $params = [];
        foreach ($vars as $idx => $var) {
          $params[$var] = intval($matches[$idx + 1]);
        }
        if ($this->runMiddlewares($middlewares)) {
          $controller->$action($params);
        }
        return true;
      }
    }
    return false;
  }

  private function isDynamic($route)
  {
    $result = preg_match_all("/\{(.*?)\}/", $route, $matches);
    if (!$result) {
      return false;
    }
    return $matches[1];
  }

  private function getRouteRegex($route)
  {
    return "/^" . preg_replace("/\{.*?\}/", "(\d+)", str_replace("/", "\\/", $route)) . "$/";
  }

  private function runMiddlewares($middlewares)
  {
    $idx = 0;

    foreach ($middlewares as $middleware) {
      $nextCalled = false;
      $middleware::handle(function () use (&$idx, &$nextCalled) {
        $idx++;
        $nextCalled = true;
      });
      if (!$nextCalled) {
        if ($idx === count($middlewares) - 1) {
          return false;
        }
        break;
      }
    }

    return true;
  }
}
